#include <stdio.h>
#include <errno.h>
#ifdef MPI
#include <mpi.h>
#endif
#include <math.h>

#include "DEFS.h"
#include "PROCESS.h"
#include "PATH.h"
#include "PROB.h"
#include "MEM.h"
#include "RNG.h"
#include "MC.h"

typedef struct mc {
int iseed;
int NMCCycles;
int Restart;
int Relaxed;
int RelaxationBlocks;
int MeasBlocks;
int MeasBlockSize;
int ConfDumpFD;
int RestartFD;
} MONTECARLO;

#ifdef MPI
  extern int MyProcessorNumber, NumberofProcessors;
#endif

static MONTECARLO MCParams;

void SetNMCCycles( int nc)
{

  if(nc < 1) {
   fprintf(stderr,"SetNMCCycles():  Number of MC Cycles must be positive\n");
   exit(1);
  }
  if(nc > MAX_NUM_MC_CYCLES) {
   fprintf(stderr,"WARNING: SetNMCCycles(): Number of MC cycles may be too large\n");
  }
  MCParams.NMCCycles = nc;
}


int ImDone()
{
  return (MCParams.NMCCycles == 0);
}



int IsRelaxed()
{
  return (MCParams.Relaxed == 1);
}



void SetRelaxFlag(int  rf)
{
  MCParams.Relaxed = rf;
}



int HowManyCyclesLeft()
{
  return (MCParams.NMCCycles);
}



void CycleDone()
{
  if((MCParams.NMCCycles)-- == 0) {
    fprintf(stderr,"Should have stopped the simulation already\n");
    exit(1);
  }
  close(GetConfDumpFD());
  close(GetRestartFD());
}



void SetRNGSeed(int iseed)
{
  if(iseed < 0) {
    fprintf(stderr,"WARNING: SetRNGSeed(): Need positive seed\n");
    fprintf(stderr,"WARNING: SetRNGSeed(): Seed set to default: %l\n", DEFAULT_SEED);
    MCParams.iseed = DEFAULT_SEED;
  } else {
    MCParams.iseed = iseed;
  }
  
}



void SetNumMeasBlocks(int mb)
{
  if(mb < 1) {
   fprintf(stderr,"SetNumMeasBlocks():  Number of measurement blocks must be positive\n");
   exit(1);
  }
  if(mb > MAX_NUM_MEAS_BLOCKS) {
   fprintf(stderr,"WARNING: SetNumMeasBlocks(): Number of measurement blocks may be too large\n");
  }
  MCParams.MeasBlocks = mb;
}



void SetMeasBlockSize(int mbs)
{
  if(mbs < 1) {
   fprintf(stderr,"WARNING: SetMeasBlockSize():  Meas block should have positive number of MC steps\n");
   fprintf(stderr,"WARNING: SetMeasBlockSize():  Set to default value: %d\n", DEFAULT_MEAS_BLOCK_SIZE);
   MCParams.MeasBlockSize = DEFAULT_MEAS_BLOCK_SIZE; 
   exit(1);
  } else {
   MCParams.MeasBlockSize = mbs;
  }
}




int HowManyMeasBlocks()
{
  return MCParams.MeasBlocks;
}



int HowBigMeasBlock()
{
  return MCParams.MeasBlockSize;
}



int HowManyRelaxBlocks()
{
  return MCParams.RelaxationBlocks;
}



void SetNumRelaxBlocks( int rb)
{
  if(rb < 0) {
   fprintf(stderr,"WARNING: SetNumRelaxBlocks():  Number of relaxation blocks cannot be negative\n");
   fprintf(stderr,"WARNING: SetNumRelaxBlocks():  Set to default value: %d\n", DEFAULT_NUM_RELAX_BLOCKS);
   MCParams.RelaxationBlocks = DEFAULT_NUM_RELAX_BLOCKS; 
   exit(1);
  } else {
   MCParams.RelaxationBlocks = rb;
  }
}



void CreateConfDumpFile(char *CDFile)
{
  int fd;
  if((fd=creat(CDFile,0644)) == EOF) {
   fprintf(stderr,"CreateConfDumpFile(): Could not create history storage file\n") ;
   exit(1) ;
  }
  MCParams.ConfDumpFD = fd;
}



int GetConfDumpFD()
{
 return MCParams.ConfDumpFD;
}




void CreateRestartFile(char *RFile)
{
  int fd;
  if((fd=creat(RFile,0644)) == EOF) {
   fprintf(stderr,"CreateRestartFile(): Could not create MC restart file\n") ;
   exit(1) ;
  }
  MCParams.RestartFD = fd;
}




int GetRestartFD()
{
 return MCParams.RestartFD;
}



int SetupMetropolisMCSimulation(char *MC_description_file)
{
  FILE *mcfile;
  int err = 0;
  int Nc;
  int mb;
  int mbs;
  int rb;
  int iseed;
  int storage;
  int i;
  int tag = 0;
#ifdef MPI
  MPI_Status status;
#endif
#ifdef MPI
  if(MyProcessorNumber == MASTER) {
#endif

  if((mcfile = fopen(MC_description_file,"r")) == NULL) {
        perror("cannot open input file for MC simulation parameters");
        exit(1);
  }
  getline(mcfile);
  fscanf(mcfile,"%d\n",&Nc);
#ifdef MPI
  }
#endif

#ifdef MPI
  MPI_Bcast(&Nc, 1, MPI_INT, 0, MPI_COMM_WORLD);
#endif
  SetNMCCycles(Nc);

#ifdef MPI
  if(MyProcessorNumber == MASTER) {
#endif

  getline(mcfile);
  fscanf(mcfile,"%d\n",&iseed);
  SetRNGSeed(iseed);
  RNG_Init(iseed);
  storage = iseed;
#ifdef MPI
  for (i = 1; i < NumberofProcessors; i++){
/**
      fscanf(mcfile,"%d\n",&storage);
**/
      storage += 101;
      MPI_Send(&storage, 1, MPI_INT, i, tag, MPI_COMM_WORLD);
  }
#endif
#ifdef MPI
  } else {
     MPI_Recv(&storage, 1, MPI_INT, 0, tag, MPI_COMM_WORLD, &status);
     SetRNGSeed(storage);
     RNG_Init(storage);
     iseed=storage;
  }
#endif


#ifdef MPI
  if(MyProcessorNumber == MASTER) {
#endif

  getline(mcfile);
  fscanf(mcfile,"%d\n",&mb);
#ifdef MPI
  }
#endif
#ifdef MPI
  MPI_Bcast(&mb, 1, MPI_INT, 0, MPI_COMM_WORLD);
#endif
  SetNumMeasBlocks(mb);

#ifdef MPI
  if(MyProcessorNumber == MASTER) {
#endif

  getline(mcfile);
  fscanf(mcfile,"%d\n",&mbs);
#ifdef MPI
  }
#endif

#ifdef MPI
  MPI_Bcast(&mbs, 1, MPI_INT, 0, MPI_COMM_WORLD);
#endif
  SetMeasBlockSize(mbs);

#ifdef MPI
  if(MyProcessorNumber == MASTER) {
#endif

  getline(mcfile);
  fscanf(mcfile,"%d\n",&rb);
#ifdef MPI
  }
#endif
#ifdef MPI
  MPI_Bcast(&rb, 1, MPI_INT, 0, MPI_COMM_WORLD);
#endif
  SetNumRelaxBlocks(rb);

#ifdef MPI
  if(MyProcessorNumber == MASTER) {
#endif

  fclose(mcfile);

#ifdef MPI
  }
#endif


  return err;
}


#ifdef BUNDLE_AR_WITH_PROB

int MetropolisUpdate
                          ( 
                          int MeasBlockSize,
                          PATH *Rates,
                          PROCESS *Params 
                          )
{
   int mcs;
   REAL *shift = NULL;
   REAL *Sshift = NULL;
   int size;
   REAL shiftPres;
   REAL shiftPrev;
   int Naccepted = 0;
   int ir;
   int NumRates = HowManyRates(Rates); 
   int Translation = ProcessSymmetry(Params,TRANSLATION);
   int Reflexion = ProcessSymmetry(Params,REFLEXION);
   REAL *AllCurrentRates = NULL;
   REAL  *AllPastRates = NULL;
   REAL  *AllFutureRates=NULL;
   REAL **Couplings = NULL;
   ProbFunc *ConditionalProbability = GetProbabilityFunction(Params);

   size = (NumRates + 1)*sizeof(REAL);
   AllCurrentRates = (REAL *) mymalloc(size,"Update(): AllCurrentRates\n");
   AllPastRates    = (REAL *) mymalloc(size,"Update(): AllPastRates\n");
   AllFutureRates    = (REAL *) mymalloc(size,"Update(): AllPastRates\n");
   shift = (REAL *) mymalloc(size,"Update(): shift\n");

   if( (Couplings = GetAllRateCouplings( Params) ) == NULL) {
     fprintf(stderr,"MetropolisUpdate():GetAllRateCouplings() failed\n");
     exit(1);
   }

   Sshift  = GetVariances(Params, NumRates);
  
   for(mcs=1;mcs<=MeasBlockSize;mcs++) {

     SetCurrentTimetoInitial(Rates);
     SetCurrentTimetoNext(Rates);


     while(!BeyondHorizon(Rates)) {
  
       if( GetCurrentTimeSlice(Rates,AllCurrentRates) != DONE) {
         fprintf(stderr,"GetCurrentTimeSlice(): Time Slice # %d\n", GetCurrentTimeIndex(Rates));
         exit(1);
       }

       if( GetPreviousTimeSlice(Rates,AllPastRates) != DONE) {
         fprintf(stderr,"GetPreviousTimeSlice(): Time Slice # %d\n", GetCurrentTimeIndex(Rates));
         exit(1);
       }

       if((!OnHorizon(Rates)) && (!Translation)) {
         if( GetNextTimeSlice(Rates,AllFutureRates) != DONE) {
           fprintf(stderr,"GetPreviousTimeSlice(): Time Slice # %d\n", GetCurrentTimeIndex(Rates));
           exit(1);
         }
       }

  
       for(ir=1;ir<=NumRates;ir++) *(shift+ir) = *(Sshift +ir) * (2.*randf()-1.);
         
       if(!OnHorizon(Rates)) {
              Naccepted += (*(ConditionalProbability))
              (NumRates, shift, AllCurrentRates, AllPastRates, AllFutureRates, Couplings);
       } else {
              Naccepted += (*(ConditionalProbability))
              (NumRates, shift, AllCurrentRates, AllPastRates, NULL, Couplings);
       }


       SetCurrentRatetoFirst(Rates);

       if(Translation) {
           while(!FinishedLastRate(Rates)) {
             ir=GetCurrentRateIndex(Rates);
             if(shift[ir] != 0.0) ShiftCurrentRateHistory(Rates, shift[ir]);
             SetCurrentRatetoNext(Rates);
           } 
       } else {
           while(!FinishedLastRate(Rates)) {
             ir=GetCurrentRateIndex(Rates);
             if(shift[ir] != 0.0) ShiftCurrentRate(Rates, shift[ir]);
				
             SetCurrentRatetoNext(Rates);
           }
       }

       SetCurrentTimetoNext(Rates);

     }
     
     if(randf()<0.5) {
        if(Reflexion) {
            FlipAllHistories(Rates);
        }
     }

   } /* end of loop over MC steps within a measurement block */

   V_Free((void *)AllCurrentRates);
   V_Free((void *)AllPastRates);
   V_Free((void *)AllFutureRates);
   V_Free((void *)shift);
   V_Free((void *)Sshift);
   M_Free((void **)Couplings,NumRates);

   return Naccepted;


} /* end of all-purpose Update() */



#else



int MetropolisUpdate
                          ( 
                          int MeasBlockSize,
                          PATH *Rates,
                          PROCESS *Params 
                          )
{
   int mcs;
   REAL prob;
   REAL shift;
   int size;
   int  RateID;
   REAL shiftPres;
   REAL  shiftPrev;
   int Naccepted = 0;
   int ir;
   int NumRates = HowManyRates(Rates); 
   int Translation = ProcessSymmetry(Params,TRANSLATION);
   int Reflexion = ProcessSymmetry(Params,REFLEXION);
   REAL *AllCurrentRates = NULL;
   REAL *AllPastRates = NULL;
   REAL *AllFutureRates=NULL;
   REAL **Couplings = NULL;
   ProbFunc *ConditionalProbability = GetProbabilityFunction(Params);


   size = (NumRates + 1)*sizeof(REAL);
   AllCurrentRates = (REAL *) mymalloc(size,"Update(): AllCurrentRates\n");
   AllPastRates    = (REAL *) mymalloc(size,"Update(): AllPastRates\n");
   AllFutureRates    = (REAL *) mymalloc(size,"Update(): AllPastRates\n");

   if( (Couplings = GetAllRateCouplings( Params) ) == NULL) {
     fprintf(stderr,"MetropolisUpdate():GetAllRateCouplings() failed\n");
     exit(1);
   }

   for(mcs=1;mcs<=MeasBlockSize;mcs++) {

     SetCurrentTimetoInitial(Rates);
     SetCurrentTimetoNext(Rates);
     while(!BeyondHorizon(Rates)) {
  
       if( GetCurrentTimeSlice(Rates,AllCurrentRates) != DONE) {
         fprintf(stderr,"GetCurrentTimeSlice(): Time Slice # %d\n", GetCurrentTimeIndex(Rates));
         exit(1);
       }
       if( GetPreviousTimeSlice(Rates,AllPastRates) != DONE) {
         fprintf(stderr,"GetPreviousTimeSlice(): Time Slice # %d\n", GetCurrentTimeIndex(Rates));
         exit(1);
       }
       if((!OnHorizon(Rates)) && (!Translation)) {
         if( GetNextTimeSlice(Rates,AllFutureRates) != DONE) {
           fprintf(stderr,"GetPreviousTimeSlice(): Time Slice # %d\n", GetCurrentTimeIndex(Rates));
           exit(1);
         }
       }
  
       SetCurrentRatetoFirst(Rates);
       while(!FinishedLastRate(Rates)) {
  
         RateID       =GetCurrentRateIndex(Rates);
         shift  = GetVariance(Params, Rates);
         shift *= (2.*randf()-1.);
         if(!OnHorizon(Rates)) {
              prob=(*(ConditionalProbability))
              (NumRates, RateID, shift,AllCurrentRates, AllPastRates, AllFutureRates, Couplings);
         } else {
              prob=(*(ConditionalProbability))
              (NumRates, RateID, shift,AllCurrentRates, AllPastRates, NULL, Couplings);
         }
  
         if(randf()<prob) {
           if(Translation) {
             ShiftCurrentRateHistory(Rates, shift);
             Naccepted++;
           } else {
             ShiftCurrentRate(Rates, shift);
             Naccepted++;
           }
         }
  
         SetCurrentRatetoNext(Rates);
  
       }
       SetCurrentTimetoNext(Rates);
     }
     
     if(randf()<0.5) {
        if(Reflexion) {
            FlipAllHistories(Rates);
        }
     }

   } /* end of loop over MC steps within a measurement block */

   V_Free((void *) AllCurrentRates);
   V_Free((void *) AllPastRates);
   V_Free((void *) AllFutureRates);
   M_Free((void **) Couplings,NumRates);

   return Naccepted;


} /* end of all-purpose Update() */

#endif
